from django.apps import AppConfig


class TaskAppConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'task_app'

    def ready(self):
        from .utils import startup
        startup(0)




import datetime
from django.utils.timezone import localtime
from django.db.models import Q
import threading
import time
from django.utils import timezone

global seconds, tasks, ack


def calculate_seconds(task):
    task.date_time = localtime(task.date_time)
    task_time = datetime.datetime(task.date_time.year, task.date_time.month, task.date_time.day, task.date_time.hour,
                                  task.date_time.minute, task.date_time.second)
    current_time = datetime.datetime.now()
    return (task_time - current_time).seconds


def create_task_stack():
    from .models import TaskQueue
    global tasks
    tasks = []
    list(map(lambda task: tasks.append(task), TaskQueue.objects.filter(status__exact='waiting')))


def complete_pending_task():
    from .models import TaskQueue
    try:
        expire_task_queryset = TaskQueue.objects.filter(Q(status__exact='waiting') & Q(date_time__lt=timezone.now()))
        for task in expire_task_queryset:
            # change article status
            task.article.status = 'inactive' if task.article.status == 'active' else 'active'
            task.article.change_status_on = None
            task.article.save()
        expire_task_queryset.update(status='completed')
        create_task_stack()
    except:
        return


def startup(wait):
    from .task import TaskEater
    global seconds, tasks, ack
    complete_pending_task()
    ack = False
    time.sleep(wait)
    while True:
        try:
            if 'TaskEater' not in list(map(lambda thread: type(thread).__name__, threading.enumerate())):
                ack = True
                latest_task = tasks[0]
                seconds = calculate_seconds(latest_task)
                worker = TaskEater('worker', latest_task)
                worker.setDaemon(True)
                worker.start()
            else:
                ack = True
                break
        except:
            break






from django.contrib import admin
from .models import TaskQueue


@admin.register(TaskQueue)
class TaskQueueAdmin(admin.ModelAdmin):
    list_display = ['id', 'status', 'article', 'date_time', 'created_on', 'updated_on']
    ordering = ['date_time']
    readonly_fields = ['id', 'article', 'status']
    list_filter = ['status']




from django.db import models
from django.dispatch import receiver
from django.db.models.signals import post_save
from blog.models import Article, AbstractCommonFields
from .choices import TASK_STATUS


class TaskQueue(AbstractCommonFields):
    status = models.CharField(max_length=10, choices=TASK_STATUS, default='waiting')
    article = models.ForeignKey(Article, on_delete=models.CASCADE)
    date_time = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return f'Task ID: {self.id} | Article: {self.article.title}'

    class Meta:
        verbose_name = 'Task'
        verbose_name_plural = 'TaskQueue'
        ordering = ['date_time', 'id']


@receiver(post_save, sender=TaskQueue)
def _post_save_receiver(sender, instance, *args, **kwargs):
    from task_app import utils
    utils.ack = False
    utils.startup(1)




import threading
import time
import datetime
from colorama import Fore


class TaskEater(threading.Thread):

    def __init__(self, name, task):
        threading.Thread.__init__(self)
        self.name = name
        self.task = task

    def run(self):
        self.timer()

    def timer(self):
        from task_app import utils
        while utils.seconds > 0:
            if not utils.ack:return
            mins, secs = divmod(utils.seconds, 60)
            timeformat = '{:02d}:{:02d}'.format(mins, secs)
            # print(timeformat)
            time.sleep(1)
            utils.seconds -= 1
        t = ChangeStatusTask(self.task)
        t.setDaemon(True)
        t.start()


class ChangeStatusTask(threading.Thread):
    def __init__(self, task):
        threading.Thread.__init__(self)
        self.task = task

    def run(self):
        time.sleep(1)
        self.change_status()

    def color_print(self, task, status):
        if status == 'active':
            print(Fore.YELLOW + f'Status Changed: {Fore.CYAN} {task.article} {Fore.YELLOW} -> {Fore.GREEN} '
                                f'{task.article.status} {Fore.YELLOW} at {Fore.BLUE} {datetime.datetime.now()} {Fore.RESET}')
        else:
            print(Fore.YELLOW + f'Status Changed: {Fore.CYAN} {task.article} {Fore.YELLOW} -> {Fore.RED} '
                                f'{task.article.status} {Fore.YELLOW} at {Fore.BLUE} {datetime.datetime.now()} {Fore.RESET}')

    def change_status(self):
        # change article status
        self.task.article.status = 'inactive' if self.task.article.status == 'active' else 'active'
        self.task.article.change_status_on = None
        self.task.article.save()
        self.color_print(self.task, self.task.article.status)
        # change task status
        self.task.status = 'completed'
        self.task.save()




# Generated by Django 4.0.2 on 2022-02-26 19:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('task_app', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskqueue',
            name='date_time',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.0.2 on 2022-02-26 18:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('blog', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='article',
            name='change_status_on',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from .serializers import TaskQueueSerializer
from .models import TaskQueue


def create_or_update_task(article):
    '''This function is to create or update task'''
    payload = {
        'article': article.id,
        'date_time': article.change_status_on
    }
    task = TaskQueue.objects.filter(article__id=article.id, status__exact='waiting').last()
    if task:
        serializer = TaskQueueSerializer(task, data=payload, partial=True)
    else:
        serializer = TaskQueueSerializer(data=payload)
    if serializer.is_valid():
        serializer.save()
        return serializer.data
    print('Task does not create.')




# Generated by Django 4.0.2 on 2022-02-26 18:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('blog', '0002_article_change_status_on'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskQueue',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True)),
                ('status', models.CharField(choices=[('waiting', 'Waiting'), ('completed', 'Completed')], default='waiting', max_length=10)),
                ('date_time', models.DateTimeField()),
                ('article', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='blog.article')),
            ],
            options={
                'verbose_name': 'Task',
                'verbose_name_plural': 'TaskQueue',
                'ordering': ['date_time', 'id'],
            },
        ),
    ]
